/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/*****************************************************************************
 * defines functions which replaces usual editor functions 
 * - to delete and insert text (additional update of command statistics and 
 *                              number editor)
 * - which involves scrolling (simultaneous scrolling in number editor)
 * - to access last/previous command relative to a command pointer
 *
 * notice:
 * Within the history tracer, a statistic is kept which holds the number of 
 * the starting and the ending line of a command. The statistic is updated 
 * whenever editing activities occur which affect the numbers of line in the
 * buffer. This book-keeping is realised by redefining the corresponding 
 * editor commands with specific functions. These functions call the original
 * functions and update the statistic according to the result of the function
 * call. 
 *
 * The History Tracer consists of two editors: one to hold the
 * commands and the other (in view mode) to hold the command numbers.  Actions
 * either take place on both editors simultaneously (scrolling etc.) or only 
 * on the command editor. To coordinate these two editors all actions which
 * change the number of lines or involve scrolling on the command editor have 
 * to be reflected in the number editor. This is done by replacing the
 * corresponding editor functions by special functions (the display of the
 * cursor has to be handled supplementary for the editor which is not the
 * current one).
 ****************************************************************************/

#include "HT.h"
#include "../exEdit/display.h"

static int HISTPTR = -1;
int ETnumlines();

/* hRight: *****************************************************************
 * move cursor right, advance to next line if end of line is reached
 ***************************************************************************/

hRight(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(exHistory->comEdit);

  if (bf->curpos < ETlen(bf, bf->curline)) {
    bf->curpos++;
    setcol(bf);
  } else if (bf->curline < ETnumlines(bf) - 1) {	
    dwnlin(exHistory->comEdit);                           /* next line */
    dwnlin(exHistory->numEdit);
  }

  putCursor(exHistory->comEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hLeft: *****************************************************************
 * move cursor left, go to previous line if beginning of line is reached
 **************************************************************************/

hLeft(editor)
     Editor editor;
{
  register  BufInfo *bf = bufinfoptr(exHistory->comEdit);

  if (bf->curpos > 0) {
    bf->curpos--;
    setcol(bf);
  } else if (bf->currow > 0) {
    uplin(exHistory->comEdit);
    uplin(exHistory->numEdit);
    beglin(exHistory->numEdit);
  }

  putCursor(exHistory->comEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hWordr: *****************************************************************
 * make cursor skip one word to the right, advance to next line if end of line
 * is reached
 ***************************************************************************/

hWordr(editor)
     Editor editor;
{
  register  BufInfo *bf = bufinfoptr(exHistory->comEdit);

  if (bf->curpos >= ETlen(bf, bf->topline + bf->currow)) {
    dwnlin(exHistory->comEdit);
    dwnlin(exHistory->numEdit);
    putCursor(exHistory->numEdit);
  }
  else
    ESwords(bf, DIR_RIGHT);

  ESsetcur(bf);

  setHistScrollbar(exHistory);
}

/* hWordl: *****************************************************************
 * make cursor skip one word to the left, advance to next line if end of line
 * is reached
 ***************************************************************************/

hWordl(editor)
     Editor editor;
{
  register  BufInfo *bf = bufinfoptr(exHistory->comEdit);

  if (bf->curpos == 0) { 
    uplin(exHistory->comEdit);
    uplin(exHistory->numEdit);
    beglin(exHistory->numEdit);
    putCursor(exHistory->numEdit);
  }
  else
    ESwords(bf, DIR_LEFT);

  ESsetcur(bf);

  setHistScrollbar(exHistory);
}

/* hDwncol: *****************************************************************
 * go down one line, preserve column
 ***************************************************************************/

hDwncol(editor)
     Editor editor;
{
  dwncol(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  dwncol(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hUpcol: *****************************************************************
 * go up one line, preserve column
 ***************************************************************************/

hUpcol(editor)
     Editor editor;
{
  upcol(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  upcol(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hBeglin: *****************************************************************
 * go to beginning of line 
 ****************************************************************************/

hBeglin(editor)
     Editor editor;
{
  beglin(exHistory->comEdit);
  putCursor(exHistory->comEdit);
}


/* hEndlin: *****************************************************************
 * go to end of line
 ****************************************************************************/

hEndlin(editor)
     Editor editor;
{
  endlin(exHistory->comEdit);
  putCursor(exHistory->comEdit);
}

/* hDwnpage: *****************************************************************
 * scroll down a screen 
 ****************************************************************************/

hDwnpage(editor)
     Editor editor;
{
  dwnpage(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  dwnpage(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hUppage: *****************************************************************
 * scroll up a screen 
 ****************************************************************************/

hUppage(editor)
     Editor editor;
{
  uppage(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  uppage(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hNewscr: *****************************************************************
 * update screen
 ****************************************************************************/

hNewscr(editor)
     Editor editor;
{
  newscr(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  newscr(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hFsearch: *****************************************************************
 * search forward: ask user for search string
 ****************************************************************************/

hFsearch(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(exHistory->comEdit), 
                   *bfe = bufinfoptr(exHistory->numEdit);
  register curline;

  curline = bf->curline;
  fsearch(editor);
  if (bf->curline != curline) {
    showline(bf, bf->curline, 0);
    ESsetcur(bf);
    showline(bfe, bf->curline, 0);
    ESsetcur(bfe);
  }

  setHistScrollbar(exHistory);
}

/* hBwsearch: *****************************************************************
 * search backward: ask user for search string
 ****************************************************************************/

hBwsearch(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(exHistory->comEdit),
                   *bfe = bufinfoptr(exHistory->numEdit);
  register curline;

  curline = bf->curline;
  bwsearch(editor);
  if (bf->curline != curline) {
    showline(bf, bf->curline, 0);
    ESsetcur(bf);
    showline(bfe, bfe->curline, 0);
    ESsetcur(bfe);
  }

  setHistScrollbar(exHistory);
}

/* hHelp: *****************************************************************
 * display the help editor if any, else notify the user that no help is 
 * available
 ****************************************************************************

 * fffhHelp(editor)
     Editor editor;
{
  help(exHistory->comEdit);
}
*/

/* hLineToTop: ***************************************************************
 * scroll text in window, so that the current line is the first line in window
 ****************************************************************************/

hLineToTop(editor)
     Editor editor;
{
  lineToTop(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  lineToTop(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hTopBuffer: **************************************************************
 * position the cursor at the beginning of the first line; if necessary 
 * the screen is scrolled, so that the first text line appears at the 
 * top of the window
 ****************************************************************************/

hTopBuffer(editor)
     Editor editor;
{
  topBuffer(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  topBuffer(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hBotBuffer: ***************************************************************
 * position the cursor in the last line; if necessary the screen is scrolled, 
 * so that the last text line appears in the middle of the window
 ****************************************************************************/

hBotBuffer(editor)
     Editor editor;
{
  botBuffer(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  botBuffer(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hGoLine: *****************************************************************
 * goto the specified physical line 
 ****************************************************************************/

hGoLine(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(exHistory->comEdit),
                   *bfe = bufinfoptr(exHistory->numEdit);
  register curline;

  curline = bf->curline;
  goLine(editor);
  if (bf->curline != curline) {
    showline(bf, bf->curline, 0);
    ESsetcur(bf);
    showline(bfe, bfe->curline, 0);
    ESsetcur(bfe);
  }

  setHistScrollbar(exHistory);
}

/* hUpScroll: *****************************************************************
 * scroll one line up
 ****************************************************************************/

hUpScroll(editor)
     Editor editor;
{
  upScroll(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  upScroll(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hDwnScroll: ***************************************************************
 * scroll down one line 
 ****************************************************************************/

hDwnScroll(editor)
     Editor editor;
{
  dwnScroll(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  dwnScroll(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hBotscr: *****************************************************************
 * position cursor to last line of the window
 ****************************************************************************/

hBotscr(editor)
     Editor editor;
{
  botscr(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  botscr(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hMidscr: *****************************************************************
 * position cursor to line in the middle of the window
 ****************************************************************************/

hMidscr(editor)
     Editor editor;
{
  midscr(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  midscr(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hTopscr: *****************************************************************
 * position cursor to first line of the window
 ****************************************************************************/

hTopscr(editor)
     Editor editor;
{
  topscr(exHistory->comEdit);
  putCursor(exHistory->comEdit);

  topscr(exHistory->numEdit);
  putCursor(exHistory->numEdit);

  setHistScrollbar(exHistory);
}

/* hFfsearch: ****************************************************************
 * search forward: use last search string if any
 ****************************************************************************/

hFfsearch(editor)
     Editor editor;
{  
  register BufInfo *bf = bufinfoptr(exHistory->comEdit),
                   *bfe = bufinfoptr(exHistory->numEdit);
  register curline;   

  curline = bf->curline;
  ffsearch(editor);
  if (bf->curline != curline) {
    showline(bf, bf->curline, 0);
    ESsetcur(bf);
    showline(bfe, bf->curline, 0);
    ESsetcur(bfe);
  }

  setHistScrollbar(exHistory);
}

/* hBbsearch: *****************************************************************
 * search backward: use last search string if any
 ****************************************************************************/

hBbsearch(editor)
     Editor editor;
{  
  register BufInfo *bf = bufinfoptr(exHistory->comEdit), 
                   *bfe = bufinfoptr(exHistory->numEdit);
  register curline;

  curline = bf->curline;
  bbsearch(editor);
  if (bf->curline != curline) {
    showline(bf, bf->curline, 0);
    ESsetcur(bf);
    showline(bfe, bfe->curline, 0);
    ESsetcur(bfe);
  }

  setHistScrollbar(exHistory);  
}

/* hFindFile: ****************************************************************
 * read a file into the editor buffer
 ****************************************************************************/

hFindFile(editor)
     Editor editor;
{
  hKillBuffer(editor);
  restoreHistory((RequestParm *) NULL);
}

/* hSaveFile: ****************************************************************
 * store the buffer contents into corresponding file if any; otherwise the
 * user is asked for a file name
 *****************************************************************************/

hSaveFile(editor)
     Editor editor;
{
  /* entry mit file namen vorbelegen ? */
  saveHistory((RequestParm *) NULL);
}

/* hWriteFile: ***************************************************************
 * ask the user for a file name in which to store the buffer contents
 ****************************************************************************/

hWriteFile(editor)
     Editor editor;
{
  saveHistory((RequestParm *) NULL);
}

/* hInsertFile: ***************************************************************
 * insert the user specified file into the buffer at the current cursor 
 * position
 ****************************************************************************/

hInsertFile(editor)
     Editor editor;
{
  restoreHistory((RequestParm *) NULL);
}

/* hEditExit: *****************************************************************
 * exit gracefully 
 ****************************************************************************/

hEditExit(editor)
     Editor editor;
{
  quitHistory((RequestParm *) NULL);
}

/* hKillBuffer: ***************************************************************
 * clear the current buffer after user confirmation
 ****************************************************************************/

hKillBuffer(editor)
     Editor editor;
{
  killBuffer(exHistory->comEdit);

  if (emptyBuffer(exHistory->comEdit)) {
    killBuffer(exHistory->numEdit);
    deleteStatistics(comStat);

    putCursor(exHistory->comEdit);
    putCursor(exHistory->numEdit);
  }

  setHistScrollbar(exHistory); 
}

/* hClearBuffer: **************************************************************
 * clear the current buffer without asking the user
 ****************************************************************************/

hClearBuffer(editor)
     Editor editor;
{
  clearBuffer(exHistory->comEdit);

  if (emptyBuffer(exHistory->comEdit)) {
    killBuffer(exHistory->numEdit);
    deleteStatistics(comStat);

    putCursor(exHistory->comEdit);
    putCursor(exHistory->numEdit);
  }

  setHistScrollbar(exHistory); 
}

/* hDelnxt: *******************************************************************
 * delete character under cursor position; if number of lines are changed 
 * update command statistics
 *****************************************************************************/

hDelnxt(editor)
     Editor editor;
{
  register  BufInfo *bf = bufinfoptr(exHistory->comEdit);
  int curLine, oldno;

  curLine = bf->topline + bf->currow;                /* current working line */
  oldno = ETnumlines(bf);

  delnxt(exHistory->comEdit);
  ESsetcur(bf);

  if ((comStat->comCount > 0) && (ETnumlines(bf) < oldno))
    updateStatistics(exHistory, comStat, curLine + 1, -1);
}

/* hDelprev: *****************************************************************
 * delete character before cursor position; if number of lines are changed 
 * update command statistics 
 ****************************************************************************/

hDelprev(editor)
     Editor editor;
{ 
  register  BufInfo *bf = bufinfoptr(exHistory->comEdit);
  int curLine, oldno;
  
  curLine = bf->topline + bf->currow;                /* current working line */
  oldno = ETnumlines(bf);

  delprev(exHistory->comEdit);
  ESsetcur(bf);

  if ((comStat->comCount > 0) && (ETnumlines(bf) < oldno))
    updateStatistics(exHistory, comStat, curLine, -1);
}

/* hFdelword: ****************************************************************
 * delete next word; if number of lines are changed update command statistics
 ****************************************************************************/

hFdelword(editor)
     Editor editor;
{
  register  BufInfo *bf = bufinfoptr(exHistory->comEdit);
  int curLine, oldno;

  curLine = bf->topline + bf->currow;                /* current working line */
  oldno = ETnumlines(bf);

  fdelword(exHistory->comEdit);
  ESsetcur(bf);

  if ((comStat->comCount > 0) && (ETnumlines(bf) < oldno))
    updateStatistics(exHistory, comStat, curLine + 1, -1);
}

/* hBdelword: ****************************************************************
 * delete last word; if number of lines are changed update command statistics
 ****************************************************************************/

hBdelword(editor)
     Editor editor;
{
  register  BufInfo *bf = bufinfoptr(exHistory->comEdit);
  int curLine, oldno;

  curLine = bf->topline + bf->currow;                /* current working line */
  oldno = ETnumlines(bf);

  bdelword(exHistory->comEdit);
  ESsetcur(bf);

  if ((comStat->comCount > 0) && (ETnumlines(bf) < oldno))
    updateStatistics(exHistory, comStat, curLine, -1);
}

/* hKrest: ******************************************************************
 * delete the rest of the line including the current position; if number of 
 * lines are changed update command statistics  
 ****************************************************************************/

hKrest(editor)
     Editor editor;
{ 
  register BufInfo *bf = bufinfoptr(exHistory->comEdit);
  int curLine, oldno;
  
  curLine = bf->topline + bf->currow;                /* current working line */
  oldno = ETnumlines(bf);

  krest(exHistory->comEdit);
  ESsetcur(bf);

  if ((comStat->comCount > 0) && (ETnumlines(bf) < oldno))
    updateStatistics(exHistory, comStat, curLine + 1, -1);
}

/* hOpenln: ******************************************************************
 * open a new line above the current line; update command statistics  
 ****************************************************************************/

hOpenln(editor)
     Editor editor;
{ 
  register BufInfo *bf = bufinfoptr(exHistory->comEdit);
  int curLine, oldno;

  autodelHist(exHistory, comStat, 1);
  
  curLine = bf->topline + bf->currow;                /* current working line */
  oldno = ETnumlines(bf);

  openln(exHistory->comEdit);
  ESsetcur(bf);

  if ((comStat->comCount > 0) && (ETnumlines(bf) > oldno))
    updateStatistics(exHistory, comStat, curLine, 1);
}

/* hOpenlnd: ******************************************************************
 * open a new line below the current line; update command statistics  
 ****************************************************************************/

hOpenlnd(editor)
     Editor editor;
{ 
  register BufInfo *bf = bufinfoptr(exHistory->comEdit);
  int curLine, oldno;

  autodelHist(exHistory, comStat, 1);
  
  curLine = bf->topline + bf->currow;                /* current working line */
  oldno = ETnumlines(bf);

  openlnd(exHistory->comEdit);
  ESsetcur(bf);

  if ((comStat->comCount > 0) && (ETnumlines(bf) > oldno))
    updateStatistics(exHistory, comStat, curLine, 1);
}

/* hNewline: ******************************************************************
 * split current line; update command statistics  
 ****************************************************************************/

hNewline(editor)
     Editor editor;
{ 
  register BufInfo *bf = bufinfoptr(exHistory->comEdit);
  int curLine, oldno;

  autodelHist(exHistory, comStat, 1);  

  curLine = bf->topline + bf->currow;                /* current working line */
  oldno = ETnumlines(bf);

  newline(exHistory->comEdit);
  ESsetcur(bf);

  if ((comStat->comCount > 0) && (ETnumlines(bf) > oldno))
    updateStatistics(exHistory, comStat, curLine, 1);
}

/* hCut: ****************************************************************
 * delete selected text; update command statistics
 *****************************************************************************/

hCut(editor)
     Editor editor;
{
  int curLine, oldno, numlines;
  register BufInfo *ebf = bufinfoptr(exHistory->comEdit);
  
  /* line in which selection ends */
  curLine = (ebf->region.d_line < ebf->region.m_line) ?
			ebf->region.m_line : ebf->region.d_line;
  oldno = ETnumlines(ebf);

  cut(exHistory->comEdit);

  numlines = ETnumlines(ebf) - oldno;
  if ((comStat->comCount > 0) && (numlines < 0)) 
    updateStatistics(exHistory, comStat, curLine, numlines);
}

/* hYank: **************************************************************
 * insert text of cut&paste buffer; update command statistics
 *****************************************************************************/

hYank(editor)
     Editor editor;
{
  int curLine, oldno, numlines;
  register BufInfo *ebf = bufinfoptr(exHistory->comEdit);
  
  curLine = ebf->topline + ebf->currow;              /* current working line */
  oldno = ETnumlines(ebf);

  yank(exHistory->comEdit);

  numlines = ETnumlines(ebf) - oldno;
  if ((comStat->comCount > 0) && (numlines > 0)) 
    updateStatistics(exHistory, comStat, curLine, numlines);
}

/* hCursorPos: ****************************************************************
 * position cursor
 *****************************************************************************/

hCursorPos(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  int curLine;
  register  BufInfo *ebf = bufinfoptr(exHistory->comEdit);
  register  BufInfo *nbf = bufinfoptr(exHistory->numEdit);
  
  curLine = ebf->topline + ebf->currow;            /* current working line */

  if (ebf->region.m_line >= 0) {
    ESdisplayRegion(ebf, False);
    reset_region(ebf);
  }

  cursorPos(ebf, event->x / exFontWidth(ebf->font), 
	         event->y / LINEHEIGHT(ebf->font));
  setDot(ebf, event->x / exFontWidth(ebf->font), 
	      event->y / LINEHEIGHT(ebf->font));

  if (curLine != ebf->topline + ebf->currow)
    cursorPos(nbf, 0, event->y / LINEHEIGHT(nbf->font));
}

/* hCursorSelect: ************************************************************
 * select text
 *****************************************************************************/

/*ARGSUSED*/
void hCursorSelect(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  register  BufInfo *ebf = bufinfoptr(exHistory->comEdit);

  cursorSelect(ebf, event->x / exFontWidth(ebf->font), 
	            event->y / LINEHEIGHT(ebf->font));
}

/* hCursorDeselect: **********************************************************
 * deselect text
 *****************************************************************************/

/*ARGSUSED*/
void hCursorDeselect(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  register  BufInfo *ebf = bufinfoptr(exHistory->comEdit);

  cursorDeselect(ebf);
}

/* hCursorCopy: *************************************************************
 * copy selected text
 ***************************************************************************/

/*ARGSUSED*/
void hCursorCopy(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  cursorCopy(bufinfoptr(exHistory->comEdit));
}

/* hCursorCut: ****************************************************************
 * delete selected text; update command statistics
 *****************************************************************************/

/*ARGSUSED*/
void hCursorCut(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  int curLine, oldno, numlines;
  register BufInfo *ebf = bufinfoptr(exHistory->comEdit);
  
  /* line in which selection ends */
  curLine = (ebf->region.d_line < ebf->region.m_line) ?
			ebf->region.m_line : ebf->region.d_line;
  oldno = ETnumlines(ebf);

  cursorCut(ebf);

  numlines = ETnumlines(ebf) - oldno;
  if ((comStat->comCount > 0) && (numlines < 0)) 
    updateStatistics(exHistory, comStat, curLine, numlines);
}

/* hCursorPaste: **************************************************************
 * insert text of cut&paste buffer; update command statistics
 *****************************************************************************/

/*ARGSUSED*/
void hCursorPaste(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  int curLine, oldno, numlines;
  register BufInfo *ebf = bufinfoptr(exHistory->comEdit);
  
  curLine = ebf->topline + ebf->currow;              /* current working line */
  oldno = ETnumlines(ebf);

  cursorPaste(ebf, event->x / exFontWidth(ebf->font), 
	           event->y / LINEHEIGHT(ebf->font));
	      /*ebf->curcol, ebf->currow);*/

  numlines = ETnumlines(ebf) - oldno;
  if ((comStat->comCount > 0) && (numlines > 0)) 
    updateStatistics(exHistory, comStat, curLine, numlines);
}

 
/* execPrevCom ***************************************************************
 * execute previous command from history (relative to internal history pointer)
 *****************************************************************************/

execPrevCom(editor)
     Editor editor;
{
  char *lcom;
  register BufInfo *bf = bufinfoptr(exHistory->numEdit);
  register BufInfo *bfe = bufinfoptr(exHistory->comEdit);

  if (comStat->comCount == 0)
    return;

  if (HISTPTR < 0)
    HISTPTR = comStat->comCount - 1;
  else if (HISTPTR - 1 >= 0)
    HISTPTR--;

  showline(bfe, comStat->comSeq[HISTPTR][0], 0);
  ESsetcur(bfe);
  showline(bf, comStat->comSeq[HISTPTR][0], 0);
  ESsetcur(bf);

  lcom = getRange(exHistory->comEdit, comStat->comSeq[HISTPTR][0], 
		  comStat->comSeq[HISTPTR][1], '\n');

  write(1, lcom, strlen(lcom));
  free(lcom);
}


/* execNextCom ***************************************************************
 * execute next command from history (relative to internal history pointer)
 *****************************************************************************/

execNextCom(editor)
     Editor editor;
{
  char *lcom;
  register BufInfo *bf = bufinfoptr(exHistory->numEdit);
  register BufInfo *bfe = bufinfoptr(exHistory->comEdit);

  if (comStat->comCount == 0)
    return;

  if (HISTPTR < 0)
    HISTPTR = comStat->comCount - 1;
  else if (HISTPTR + 1 <= comStat->comCount - 1)
    HISTPTR++;

  showline(bfe, comStat->comSeq[HISTPTR][0], 0);
  ESsetcur(bfe);
  showline(bf, comStat->comSeq[HISTPTR][0], 0);
  ESsetcur(bf);

  lcom = getRange(exHistory->comEdit, comStat->comSeq[HISTPTR][0], 
		  comStat->comSeq[HISTPTR][1], '\n');

  write(1, lcom, strlen(lcom));
  free(lcom);
}


/* execLastCom ***************************************************************
 * execute last command from history
 ****************************************************************************/

execLastCom(editor)
     Editor editor;
{
  char *lcom;
  register BufInfo *bf = bufinfoptr(exHistory->numEdit);
  register BufInfo *bfe = bufinfoptr(exHistory->comEdit);

  if (comStat->comCount == 0)
    return;

  HISTPTR = comStat->comCount - 1;

  showline(bfe, comStat->comSeq[HISTPTR][0], 0);
  ESsetcur(bfe);
  showline(bf, comStat->comSeq[HISTPTR][0], 0);
  ESsetcur(bf);

  lcom = getRange(exHistory->comEdit, comStat->comSeq[HISTPTR][0], 
		  comStat->comSeq[HISTPTR][1], '\n');

  write(1, lcom, strlen(lcom));
  free(lcom); 
}

/* execThisCom ***************************************************************
 * set internal history pointer and execute current command from history
 ****************************************************************************/

execThisCom(editor)
     Editor editor;
{
  char *lcom;
  register BufInfo *bf = bufinfoptr(exHistory->numEdit);
  register BufInfo *bfe = bufinfoptr(exHistory->comEdit);

  if (comStat->comCount == 0)
    return;

  setHistPtr(editor);

  lcom = getRange(exHistory->comEdit, comStat->comSeq[HISTPTR][0], 
		  comStat->comSeq[HISTPTR][1], '\n');

  write(1, lcom, strlen(lcom));
  free(lcom); 
}

/* setHistPtr ***************************************************************
 * set internal history pointer
 ****************************************************************************/

setHistPtr(editor)
     Editor editor;
{
  register i;
  register BufInfo *bfe = bufinfoptr(exHistory->comEdit);
  int curLine;

  curLine = bfe->topline + bfe->currow;

  if (comStat->comCount == 0)
    return;

  for (i = comStat->comCount - 1; i >= 0; i--) {
    if (comStat->comSeq[i][0] <= curLine && comStat->comSeq[i][1] >= curLine){ 
	                                     /* curLine in current command */
      HISTPTR = i;
      return;
    }
  }

  HISTPTR = comStat->comCount - 1;
  editorMsg(exHistory->comEdit, "No command at this cursor position");
}

/* joinWithNextCom **********************************************************
 * join current command with next command 
 ****************************************************************************/

joinWithNextCom(editor)
     Editor editor;
{
  register i;
  register BufInfo *bfe = bufinfoptr(exHistory->comEdit);
  int curLine;

  curLine = bfe->topline + bfe->currow;

  if (comStat->comCount <= 1)
    return;

  if ((i = getComIndex(curLine, comStat)) >= 0 && 
      (comStat->comSeq[i + 1][0] > 0)) {
    updateStatistics(exHistory, comStat, comStat->comSeq[i + 1][0], -1);
    updateStatistics(exHistory, comStat, comStat->comSeq[i][1], 1);

    if (HISTPTR > i)
      HISTPTR--;
  }
}

/* insertCom ****************************************************************
 * insert new command behind current command
 ****************************************************************************/

insertCom(editor)
     Editor editor;
{
  register i, index;
  register BufInfo *bfe = bufinfoptr(exHistory->comEdit);
  int curLine;

  autodelHist(exHistory, comStat, 1);

  curLine = bfe->topline + bfe->currow;
  index = getComIndex(curLine, comStat);             /* find current command */
  curLine = comStat->comSeq[index][1];           /* get last line of command */

  i = comStat->comCount - 1;   
  while (i > index) {       /* adapt complete commands stored behind changes */
      comStat->comSeq[i+1][0] = comStat->comSeq[i][0] + 1;
      comStat->comSeq[i+1][1] = comStat->comSeq[i][1] + 1;
      i--;
    }

  comStat->comSeq[i+1][0] = curLine + 1;
  openlnd(exHistory->comEdit);    /* insert new line behind current command */
  openlnd(exHistory->numEdit);
  editorInsert(exHistory->numEdit, " +:");
  comStat->comSeq[i+1][1] = curLine + 1;
  comStat->comCount++;

  if (HISTPTR > index)
    HISTPTR++;
}

/* adjustHistPtr ************************************************************
 * adjust internal history pointer to internal changes to command statistics,
 * e.g. if autodelete is performed
 ***************************************************************************/

adjustHistPtr(diff)
     int diff;
{
  if (HISTPTR == -1)
    return;

  HISTPTR += diff;

  if (HISTPTR < 0)
    HISTPTR = 0;
  else if (HISTPTR > comStat->comCount - 1)
    HISTPTR = comStat->comCount - 1;
}
